/*
 * File:        main.c
 * Author:      Adam Chung, Ashley Xue, Jessica Chen
 * Target PIC:  PIC32MX250F128B
 */

////////////////////////////////////
// clock AND protoThreads configure!
// You MUST check this file!
#include "config.h"
// threading library
#include "pt_cornell_1_2_1.h"

#include "servo.h"
#include "i2cIMU.h"

////////////////////////////////////
// graphics libraries
#include "tft_master.h"
#include "tft_gfx.h"

// === thread structures ============================================
// thread control structs
static struct pt pt_control, pt_motors;

unsigned char flag = 0;

signed short initPitch, initRoll, initYaw;
signed short eulerPitch, eulerRoll, eulerYaw;
signed short pitchDiff, rollDiff, yawDiff, prevPitchDiff, prevRollDiff;

// string buffer
char buffer[60];

// === Update LCD Thread =================================================
static PT_THREAD (protothread_lcd(struct pt *pt))
{
    PT_BEGIN(pt);        
    while(1) {
        PT_YIELD_TIME_msec(50);
        if (!flag) { 
          // sets base values for the relative position
          initPitch = i2c1ConsecutiveRead(EULER_PITCH_LSB_ADDR, 2);
          initYaw = i2c1ConsecutiveRead(EULER_YAW_LSB_ADDR, 2);
          initRoll = i2c1ConsecutiveRead(EULER_ROLL_LSB_ADDR, 2);
          flag = 1;
          // reset the motor positions to center
          rotate(0, 1);
          rotate(7, 2);
        }
        
        // update the relative position euler angles
        eulerPitch = i2c1ConsecutiveRead(EULER_PITCH_LSB_ADDR, 2);
        eulerYaw = i2c1ConsecutiveRead(EULER_YAW_LSB_ADDR, 2);
        eulerRoll = i2c1ConsecutiveRead(EULER_ROLL_LSB_ADDR, 2);
        
        // display current relative euler angles
        tft_fillRoundRect(10, 10, 200, 14, 1, ILI9340_BLACK);
        tft_setCursor(10, 10);
        tft_setTextColor(ILI9340_CYAN);
        tft_setTextSize(1);
        sprintf(buffer, "Init: %f, Pitch: %f", initPitch / 16.0, eulerPitch / 16.0);
        tft_writeString(buffer);
        
        tft_fillRoundRect(10, 30, 200, 14, 1, ILI9340_BLACK);
        tft_setCursor(10, 30);
        tft_setTextColor(ILI9340_CYAN);
        tft_setTextSize(1);
        sprintf(buffer, "Init: %f, Yaw: %f", initYaw / 16.0, eulerYaw / 16.0);
        tft_writeString(buffer);
        
        tft_fillRoundRect(10, 50, 200, 14, 1, ILI9340_BLACK);
        tft_setCursor(10, 50);
        tft_setTextColor(ILI9340_CYAN);
        tft_setTextSize(1);
        sprintf(buffer, "Init: %f, Roll: %f", initRoll / 16.0, eulerRoll / 16.0);
        tft_writeString(buffer);

        // divide by 16 to account for sensitivity of 16 LSB
        pitchDiff = eulerPitch/16.0 - initPitch/16.0;
        yawDiff = eulerYaw/16.0 - initYaw/16.0;
        rollDiff = eulerRoll/16.0 - initRoll/16.0;
        
        // display errors in pitch and roll
        tft_fillRoundRect(10, 70, 200, 14, 1, ILI9340_BLACK);
        tft_setCursor(10, 70);
        tft_setTextColor(ILI9340_CYAN);
        tft_setTextSize(1);
        sprintf(buffer, "Pitch diff: %f", eulerPitch/16.0 - initPitch/16.0);
        tft_writeString(buffer);
        
        tft_fillRoundRect(10, 90, 200, 14, 1, ILI9340_BLACK);
        tft_setCursor(10, 90);
        tft_setTextColor(ILI9340_CYAN);
        tft_setTextSize(1);
        sprintf(buffer, "Roll diff: %f", eulerRoll/16.0 - initRoll/16.0);
        tft_writeString(buffer);
        
        // only update the motors if the base has moved
        if (pitchDiff != prevPitchDiff)
            rotate(-pitchDiff, 1);
     
        if (rollDiff != prevRollDiff)
            rotate(-rollDiff + 7, 2);

        // update previous pitch and roll differences  
        prevPitchDiff = pitchDiff;
        prevRollDiff = rollDiff;

    } // END WHILE(1)
    PT_END(pt);
  } // lcd thread

// === Motor Thread ===============================================
static PT_THREAD (protothread_motor(struct pt *pt))
{
    PT_BEGIN(pt);        
    while(1) {
      // calibrates the motors to start in the center
      rotate(0, 1);
      rotate(7, 2);
    }
    PT_END(pt);
}
// === Main  ======================================================
void main(void) {  
  ANSELA = 0; ANSELB = 0; 

  // === Timer 3: Setting the period of the PWM ===
  // 1. Open Timer 3 with a period of 50,000 with a prescalar of 16
  //    to exhibit a PWM frequency of 50Hz.
  //      a. One period should be 40MHz / 50Hz = 800,000 timer counts
  //      b. Timer overflows at around 60,000, so we can use a prescalar
  //         of 16 because it gives a timer value less than 60000; 
  //         800,000/16 = 50,000.
  // 2. Enable Output Compare 1 and 2 with Timer 3 as a source
  //    in order to output the PWM waves to the servos
  OpenTimer3(T3_ON | T3_SOURCE_INT | T3_PS_1_16, 50000);
  OpenOC1(OC_ON | OC_PWM_FAULT_PIN_ENABLE | OC_TIMER3_SRC, 0, 0);    
  OpenOC2(OC_ON | OC_PWM_FAULT_PIN_ENABLE | OC_TIMER3_SRC, 0, 0);
  // Set pins 2 and 14 as outputs for the Output Compares
  // OC1 - Pin 2: Updated c to adjust RPM of the fan
  // OC2 - Pin 14: Current RPM of the fan
  PPSOutput(1, RPA0, OC1);
  PPSOutput(2, RPB5, OC2);
    
  // SDA = Pin 18
  // SCL = Pin 17
  // I2C1CONbits.SCLREL = 1;
  // I2C1CONbits.DISSLW = 1; // Slew Rate disabled for 100KHz mode
  // I2C1CONbits.ON = 1;     // Enable I2C1 module
  // I2C1BRG = 93;
  // calculated as follows: BRG = (Fpb / 2 / baudrate) - 2.
  // OPERATION_MODE_IMUPLUS chooses the sensor to operate as an IMU
  OpenI2C1(I2C_ON | I2C_SLW_DIS, 93);
  configIMU(OPERATION_MODE_IMUPLUS);

  // === setup system wide interrupts  ========
  INTEnableSystemMultiVectoredInt();

  // === config threads ==========
  // turns OFF UART support and debugger pin, unless defines are set
  PT_setup();
  
  // init the threads
  PT_INIT(&pt_control);
  PT_INIT(&pt_motors);
 
  // init the display
  tft_init_hw();
  tft_begin();
  tft_fillScreen(ILI9340_BLACK);
   
  // round-robin scheduler for threads
  while (1){
    PT_SCHEDULE(protothread_lcd(&pt_control));
    // uncomment this and comment the first thread to calibrate motors
    // PT_SCHEDULE(protothread_motor(&pt_motors));
  }
} // main
// === end  ======================================================

/*****************************************************************************
 * Initialization of the IMU
 *
 * This function brings the IMU into "mode" operation mode to allow for access 
 * of predefined register values. Registers are: 
 *****************************************************************************/
void configIMU(unsigned char mode) {
    // start the transmission sequence
    StartI2C1();
    IdleI2C1();

    // write the device address with R/W as 0
    MasterWriteI2C1(DEVICE_ADDRESS << 1); 
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return; 

    // write the address of the operation mode register
    MasterWriteI2C1(OPR_MODE_ADDR); 
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return; 
    
    // we pick the IMU mode to get the fusion data
    // this allows us to get the degrees relative to the IMU for pitch, yaw,
    // and roll
    // this writes mode to the operation register
    MasterWriteI2C1(mode); // pick mode
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return; 
    
    // stop the transmission sequence
    StopI2C1();
    IdleI2C1();
}

/*****************************************************************************
 * Read from I2C
 * 
 * This allows the PIC to read values from the IMU registers.
 * Arguments:
 * addr - this is the register address you want to read from in the IMU
 *****************************************************************************/
signed char i2c1Read(unsigned char addr){
    unsigned char data;

    // start the transmission sequence 
    StartI2C1();
    IdleI2C1();
    // write the device address with R/W as 0
    MasterWriteI2C1(DEVICE_ADDRESS << 1);
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return; 

    // write the address of the register you want to read from
    MasterWriteI2C1(addr);
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return; 
    
    //initiate repeated start 
    StartI2C1();
    IdleI2C1();
    // rewrite the device address with R/W as 1 this time
    MasterWriteI2C1((DEVICE_ADDRESS << 1) | 1); // Issue Read Command
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return; 

    // read the data from the specified register and nack to end transmission
    data = MasterReadI2C1(); 
    NotAckI2C1();
    // stop the transmission sequence
    StopI2C1();
    IdleI2C1();

    return data;
}

/*****************************************************************************
 * Consecutive Read from I2C
 * 
 * This allows the PIC to read from consecutive registers from the IMU.
 * Arguments:
 * startReg - this is the first register you want to read from
 * num - this is the number of consecutive registers you want to read from
 *****************************************************************************/
signed short i2c1ConsecutiveRead(unsigned char startReg, unsigned char num) {
    // the first and second bytes read
    unsigned char lsb;
    signed char msb;
    // start the transmission sequence
    StartI2C1();
    IdleI2C1();
    // write the device address with R/W as 0
    MasterWriteI2C1(DEVICE_ADDRESS << 1); 
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return; 

    // write the first register you want to read from
    MasterWriteI2C1(startReg); 
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return; 

    //initiate repeated start 
    StartI2C1();
    IdleI2C1();
    // write the device address again but this time with R/W as 1
    MasterWriteI2C1((DEVICE_ADDRESS << 1) | 1); 
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return; 

    // read the data for the first byte and send ack to slave to read consecutively
    lsb = MasterReadI2C1(); 
    AckI2C1();
    IdleI2C1();
    // read the data for the second byte and send nack to end transmission
    msb = MasterReadI2C1(); 
    NotAckI2C1();
    // stop the transmission sequence
    StopI2C1();
    IdleI2C1();
    // return the two bytes in the form of a signed short
    return (msb << 8)| lsb;
}
